# !/usr/bin/python
# coding=utf-8
try:
	import pymel.core as pm
except ImportError as error:
	print (__file__, error)


class Macro():
	"""Assign macro functions to hotkeys.

	ex. call:
	class MSlots(Macro):
		'''A class that inherits from `Macro` and holds any macro functions.
		'''
		@staticmethod
		def m_back_face_culling():
			'''Toggle Back-Face Culling.
			'''
			sel = pm.ls(selection=True)
			if sel:
				currentPanel = getPanel(withFocus=True)
				state = pm.polyOptions(sel, query=True, wireBackCulling=True)[0]

				if not state:
					pm.polyOptions(sel, gl=True, wireBackCulling=True)
					Macros.setWireframeOnShadedOption(currentPanel, 0)
					pm.inViewMessage(statusMessage="Back-Face Culling is now <hl>OFF</hl>.>", pos='topCenter', fade=True)
				else:
					pm.polyOptions(sel, gl=True, backCulling=True)
					Macros.setWireframeOnShadedOption(currentPanel, 1)
					pm.inViewMessage(statusMessage="Back-Face Culling is now <hl>ON</hl>.", pos='topCenter', fade=True)
			else:
				print(" Warning: Nothing selected. ")

	#call the `setMacros` function to set a macro for functions you defined in `MSlots`.
	MSlots.setMacros(macros={
		'm_back_face_culling': {'k':'1', 'cat':'Display'},
	}
	"""
	@classmethod
	def setMacros(cls, macros={}):
		'''Extends setMacro to accept a dictionary.

		:Parameters:
			macros (dict) = Command names as keys, with dict values containing any keyword args for 'setMacro'. 
			ex. {'m_group': {'k':'ctl+g', 'cat':'Edit'}}
		'''
		for name, kwargs in macros.items():
			cls.setMacro(name, **kwargs)


	@classmethod
	def setMacro(cls, name, k=None, cat=None, ann=None, default=False, deleteExisting=True):
		'''Sets a default runtime command with a keyboard shortcut.

		:Parameters:
			name (str) = The command name you provide must be unique. (alphanumeric characters, or underscores)
			cat (str) = catagory - Category for the command.
			ann (str) = annotation - Description of the command.
			k (str) = keyShortcut - Specify what key is being set.
						key modifier values are set by adding a '+' between chars. ie. 'sht+z'.
						modifiers:
							alt, ctl, sht
						additional valid keywords are:
							Up, Down, Right, Left,
							Home, End, Page_Up, Page_Down, Insert
							Return, Space
							F1 to F12
							Tab (Will only work when modifiers are specified)
							Delete, Backspace (Will only work when modifiers are specified)
			default (bool) = Indicate that this run time command is a default command. Default run time commands will not be saved to preferences.
			deleteExisting = Delete any existing (non-default) runtime commands of the given name.
		'''
		command = f"if 'm_slots' not in globals(): from {cls.__module__} import {cls.__name__}; global m_slots; m_slots = {cls.__name__}();\nm_slots.{name}();"

		if not ann: #if no ann is given, try using the method's docstring.
			method = getattr(cls, name)
			ann = method.__doc__.split('\n')[0] #use only the first line.

		if pm.runTimeCommand(name, exists=True):
			if pm.runTimeCommand(name, query=True, default=True):
				return #can not delete default runtime commands.
			elif deleteExisting:#delete any existing (non-default) runtime commands of that name.
				pm.runTimeCommand(name, edit=True, delete=True)

		try: #set runTimeCommand
			pm.runTimeCommand(
				name,
				annotation=ann,
				category=cat,
				command=command,
				default=default,
			)
		except RuntimeError as error:
			print ('# Error: {}: {} #'.format(__file__, error))
			return error

		#set command
		nameCommand = pm.nameCommand(
				'{0}Command'.format(name),
				annotation=ann,
				command=name,
		)

		#set hotkey
		#modifiers
		ctl=False; alt=False; sht=False
		for char in k.split('+'):
			if char=='ctl':
				ctl = True
			elif char=='alt':
				alt = True
			elif char=='sht':
				sht = True
			else:
				key = char

		# print(name, char, ctl, alt, sht)
		pm.hotkey(keyShortcut=key, name=nameCommand, ctl=ctl, alt=alt, sht=sht) #set only the key press.

# -----------------------------------------------









# -----------------------------------------------
# Notes
# -----------------------------------------------

'''
#create wrapper
mel.createMelWrapper(method)

#set command
pm.nameCommand('name', annotation='', command=<>)
pm.hotkey(key='1', altModifier=True, name='name')


#clear keyboard shortcut
pm.hotkey(keyShortcut=key, name='', releaseName='', ctl=ctl, alt=alt, sht=sht) #unset the key press name and releaseName.


#query runTimeCommand
if pm.runTimeCommand('name', exists=True):


#delete runTimeCommand
pm.runTimeCommand('name', edit=True, delete=True)


#set runTimeCommand
pm.runTimeCommand(
			'name',
			annotation=string,
			category=string,
			categoryArray,
			command=script,
			commandArray,
			commandLanguage=string,
			default=boolean,
			defaultCommandArray,
			delete,
			exists,
			hotkeyCtx=string,
			image=string,
			keywords=string,
			annotation=string,
			longAnnotation=string,
			numberOfCommands,
			numberOfDefaultCommands,
			numberOfUserCommands,
			plugin=string,
			save,
			showInHotkeyEditor=boolean,
			tags=string,
			userCommandArray,
)

-annotation(-ann) string createqueryedit 
		Description of the command.

-category(-cat) string createqueryedit	
		Category for the command.

-categoryArray(-caa) query			
		Return all the run time command categories.

-command(-c) script createqueryedit		
		Command to be executed when runTimeCommand is invoked.

-commandArray(-ca) query				
		Returns an string array containing the names of all the run time commands.

-commandLanguage(-cl) string createqueryedit
		In edit or create mode, this flag allows the caller to choose a scripting language for a command passed to the "-command" flag. If this flag is not specified, then the callback will be assumed to be in the language from which the runTimeCommand command was called. In query mode, the language for this runTimeCommand is returned. The possible values are "mel" or "python".

-default(-d) boolean createquery 		
		Indicate that this run time command is a default command. Default run time commands will not be saved to preferences.

-defaultCommandArray(-dca) query				
		Returns an string array containing the names of all the default run time commands.

-delete(-del) edit 				
		Delete the specified user run time command.

-exists(-ex) create 				
		Returns true|false depending upon whether the specified object exists. Other flags are ignored.

-hotkeyCtx(-hc)	string createqueryedit 	
		hotkey Context for the command.

-image(-i) string createqueryedit 	
		Image filename for the command.

-keywords(-k) string createqueryedit		
		Keywords for the command. Used for searching for commands in Type To Find. When multiple keywords, use ; as a separator. (Example: "keyword1;keyword2")

-annotation(-annotation) string createqueryedit		
		Label for the command.

-longAnnotation(-la) string createqueryedit	
		Extensive, multi-line description of the command. This will show up in Type To Finds more info page in addition to the annotation.

-numberOfCommands(-nc) query			
		Return the number of run time commands.

-numberOfDefaultCommands(-ndc) query			
		Return the number of default run time commands.

-numberOfUserCommands(-nuc)	query			
		Return the number of user run time commands.

-plugin(-p)	string createqueryedit			
		Name of the plugin this command requires to be loaded. This flag wraps the script provided into a safety check and automatically loads the plugin referenced on execution if it hasn't been loaded. If the plugin fails to load, the command won't be executed.

-save(-s) edit 							
		Save all the user run time commands.

-showInHotkeyEditor(-she) boolean createqueryedit		
		Indicate that this run time command should be shown in the Hotkey Editor. Default value is true.

-tags(-t) string createqueryedit	
		Tags for the command. Used for grouping commands in Type To Find. When more than one tag, use ; as a separator. (Example: "tag1;tag2")

-userCommandArray(-uca)	query			
		Returns an string array containing the names of all the user run time commands.
'''



# Depricated: -------------------------------------------------------------------------------------------------------


## command = self.formatSource(name, removeTabs=1) #remove 1 tab space.
# def formatSource(self, cmd, removeTabs=0):
# 		'''Return the text of the source code for an object.
# 		The source code is returned as a single string.
# 		Removes lines containing '@' or 'def ' ie. @staticmethod.

# 		:Parameters:
# 			cmd = module, class, method, function, traceback, frame, or code object.
# 			removeTabs (int) = remove x instances of '\t' from each line.

# 		:Return:
# 			A Multi-line string.
# 		'''
# 		from inspect import getsource
# 		source = getsource(getattr(Macros, cmd))

# 		l = [s.replace('\t', '', removeTabs) for s in source.split('\n') if s and not '@' in s]
# 		call = [s.replace('\t', '', removeTabs).lstrip('def ').rstrip(':') for s in source.split('\n') if 'def ' in s]
# 		return '\n'.join(l)+'\n\n'+call[0]



		# string $image_plane[] = `ls -exactType imagePlane`;
		# for ($object in $image_plane){
		# 	if (`getAttr ($object+".displayMode")` != 2){
		# 		setAttr ($object+".displayMode") 2;
		# 		grid -toggle 1;
		# 		inViewMessage -statusMessage "Grid is now <hl>ON</hl>.\\n<hl>F1</hl>"  -fade -position topCenter;
		# 	}else{
		# 		setAttr ($object+".displayMode") 0;
		# 		grid -toggle 0;
		# 		inViewMessage -statusMessage "Grid is now <hl>OFF</hl>.\\n<hl>F1</hl>"  -fade -position topCenter;
		# 	}
		# }


# 		global int $toggleFrame_;

# 		string $selection[] = `ls -selection`;

# 		$mode = `selectMode -query -component`;
# 		$maskVertex = `selectType -query -vertex`;
# 		$maskEdge = `selectType -query -edge`;
# 		$maskFacet = `selectType -query -facet`;

# 		if (size($selection)==0)
# 			{
# 			viewFit -allObjects;
# 			}
			
# 		if ($mode==1 && $maskVertex==1 && size($selection)!=0)
# 			{
# 			if (size($selection)>1)
# 				{
# 				if ($toggleFrame_ == !1)
# 					{
# 					viewFit -fitFactor .65;
# 					$toggleFrame_ = 1;
# 					print ("frame vertices "+$toggleFrame_+"\\n");
# 					}
# 				else
# 					{
# 					viewFit -fitFactor .10;
# 					//viewSet -previousView;
# 					$toggleFrame_ = 0;
# 					print ("frame vertices "+$toggleFrame_+"\\n");
# 					}
# 				}
# 			else
# 				{
# 				if ($toggleFrame_ == !1)
# 					{
# 					viewFit -fitFactor .15;
# 					$toggleFrame_ = 1;
# 					print ("frame vertex "+$toggleFrame_+"\\n");
# 					}
# 				else
# 					{
# 					viewFit -fitFactor .01;
# 					//viewSet -previousView;
# 					$toggleFrame_ = 0;
# 					print ("frame vertex "+$toggleFrame_+"\\n");
# 					}
# 				}
# 			}
# 		if ($mode==1 && $maskEdge==1 && size($selection)!=0)
# 			{
# 			if ($toggleFrame_ == !1)
# 				{
# 				viewFit -fitFactor .3;
# 				$toggleFrame_ = 1;
# 				print ("frame edge "+$toggleFrame_+"\\n");
# 				}
# 			else
# 				{
# 				viewFit -fitFactor .9;
# 				//viewSet -previousView;
# 				$toggleFrame_ = 0;
# 				print ("frame edge "+$toggleFrame_+"\\n");
# 				}
# 			}
# 		if ($mode==1 && $maskFacet==1)
# 			{
# 			if ($toggleFrame_ == !1)
# 				{
# 				viewFit -fitFactor .9;
# 				$toggleFrame_ = 1;
# 				print ("frame facet "+$toggleFrame_+"\\n");
# 				}
# 			else
# 				{
# 				viewFit -fitFactor .45;
# 				//viewSet -previousView;
# 				$toggleFrame_ = 0;
# 				print ("frame facet "+$toggleFrame_+"\\n");
# 				}
# 			}
# 		else if ($mode==0  && size($selection)!=0)
# 			{
# 			if ($toggleFrame_ == !1)
# 				{
# 				viewFit -fitFactor .99;
# 				$toggleFrame_ = 1;
# 				print ("frame object "+$toggleFrame_+"\\n");
# 				}
# 			else
# 				{
# 				viewFit -fitFactor .65;
# 				//viewSet -previousView;
# 				$toggleFrame_ = 0;
# 				print ("frame object "+$toggleFrame_+"\\n");
# 				}
# 			}




		# string $currentPanel = `getPanel -withFocus`;
		# $mode = `displayPref -query -wireframeOnShadedActive`;

		# if ($mode=="none")
		# 	{
		# 	displayPref -wireframeOnShadedActive "reduced";
		# 	setWireframeOnShadedOption 1 $currentPanel;
		# 	inViewMessage -statusMessage "<hl>Wireframe-on-selection</hl> is now <hl>Full</hl>.\\n<hl>3</hl>"  -fade -position topCenter;
		# 	}
		# if ($mode=="reduced")
		# 	{
		# 	displayPref -wireframeOnShadedActive "full";
		# 	setWireframeOnShadedOption 0 $currentPanel;
		# 	inViewMessage -statusMessage "<hl>Wireframe-on-selection</hl> is now <hl>Reduced</hl>.\\n<hl>3</hl>"  -fade -position topCenter;
		# 	}
		# if ($mode=="full")
		# 	{
		# 	displayPref -wireframeOnShadedActive "none";
		# 	setWireframeOnShadedOption 0 $currentPanel;
		# 	inViewMessage -statusMessage "<hl>Wireframe-on-selection</hl> is now <hl>OFF</hl>.\\n<hl>3</hl>" -fade -position topCenter;
		# 	}


	# //xray all except selected
	# string $scene[] = `ls -visible -flatten -dag -noIntermediate -type surfaceShape`;
	# string $selection[] = `ls -selection -dagObjects -shapes`;
	# for ($object in $scene)
	# 	{
	# 	if (!stringArrayContains ($object, $selection))
	# 		{
	# 		int $state[] = `displaySurface -query -xRay $object`;
	# 		displaySurface -xRay ( !$state[0] ) $object;
	# 		}
	# 	}


		# mel.eval('''
		# string $currentPanel = `getPanel -withFocus`;
		# string $state = `modelEditor -query -displayAppearance $currentPanel`;
		# string $displayTextures = `modelEditor -query -displayTextures $currentPanel`;
		# if(`modelEditor -exists $currentPanel`)
		#   {
		# 	if($state != "wireframe" && $displayTextures == false)
		# 	  {
		# 		modelEditor -edit -displayAppearance smoothShaded -activeOnly false -displayTextures true $currentPanel;
		# 		inViewMessage -statusMessage "modelEditor -smoothShaded <hl>true</hl> -displayTextures <hl>true</hl>.\\n<hl>5</hl>"  -fade -position topCenter;
		# 		}
		# 	if($state == "wireframe" && $displayTextures == true)
		# 	  {
		# 		modelEditor -edit -displayAppearance smoothShaded -activeOnly false -displayTextures false $currentPanel;
		# 		inViewMessage -statusMessage "modelEditor -smoothShaded <hl>true</hl> -displayTextures <hl>false</hl>.\\n<hl>5</hl>"  -fade -position topCenter;
		# 		}
		# 	if($state != "wireframe" && $displayTextures == true)
		# 	  {
		# 		modelEditor -edit -displayAppearance wireframe -activeOnly false $currentPanel;
		# 		inViewMessage -statusMessage "modelEditor -wireframe <hl>true</hl>.\\n<hl>5</hl>"  -fade -position topCenter;
		# 		}
		# 	}
		# ''')


# SelectMultiComponentMask;
# inViewMessage -statusMessage "<hl>Multi-Component Selection Mode</hl>\\n Mask is now <hl>ON</hl>.\\n<hl>F4</hl>"  -fade -position topCenter;



		# //paste and then re-name object removing keyword 'pasted'
		# cutCopyPaste "paste";
		# {
		# string $pasted[] = `ls "pasted__*"`;
		# string $object;
		# for ( $object in $pasted )
		# {
		# string $elements[];
		# // The values returned by ls may be full or partial dag
		# // paths - when renaming we only want the actual
		# // object name so strip off the leading dag path.
		# //
		# tokenize( $object, "|", $elements );
		# string $stripped = $elements[ `size $elements` - 1 ];
		# // Remove the 'pasted__' suffix from the name
		# //
		# $stripped = `substitute "pasted__" $stripped ""`;
		# // When renaming a transform its shape will automatically be
		# // be renamed as well. Use catchQuiet here to ignore errors
		# // when trying to rename the child shape a second time.
		# // 
		# catchQuiet(`evalEcho("rename " + $object + " " + $stripped)`);
		# }
		# };
		# //alternative: edit the cutCopyPaste.mel
		# //REMOVE the line "-renameAll" so the sub-nodes won't get renamed at all
		# //REMOVE the -renamingPrefix "paste_" line
		# //and instead write the line -defaultNamespace


# $mode = `selectMode -query -component`;
# if ($mode==0)
#   {
# 	changeSelectMode -component;
# 	}

# $maskVertex = `selectType -query -vertex`;
# $maskEdge = `selectType -query -edge`;
# $maskFacet = `selectType -query -facet`;

# if ($maskEdge==0 && $maskFacet==1)
# 	{
# 	selectType -vertex true;
# 	inViewMessage -statusMessage "<hl>Vertex</hl> Mask is now <hl>ON</hl>.\\n<hl>F4</hl>"  -fade -position topCenter;
# 	}
# if ($maskVertex==1 && $maskFacet==0)
# 	{
# 	selectType -edge true;
# 	inViewMessage -statusMessage "<hl>Edge</hl> Mask is now <hl>ON</hl>.\\n<hl>F4</hl>"  -fade -position topCenter;
# 	}
# if ($maskVertex==0 && $maskEdge==1)
# 	{
# 	selectType -facet true;
# 	inViewMessage -statusMessage "<hl>Facet</hl> Mask is now <hl>ON</hl>.\\n<hl>F4</hl>"  -fade -position topCenter;
# 	}


		# // added this expression to fix 'toggleMainMenubar function not found' error
		# if (`menu -q -ni MayaWindow|HotBoxControlsMenu` == 0) {setParent -m MayaWindow|HotBoxControlsMenu;source HotboxControlsMenu;};

		# //toggle panel menus
		# string $panels[] = `getPanel -allPanels`;
		# int $state = `panel -query -menuBarVisible $panels[0]`;
		# for ($panel in $panels)
		# {
		# 	// int $state = `panel -query -menuBarVisible $panel`;
		# 	panel -edit -menuBarVisible (!$state) $panel;
		# }
		# //toggle main menubar
		# toggleMainMenubar (!$state);

		# //toggle titlebar
		# window -edit -titleBar (!$state) $gMainWindow;

		# // //toggle fullscreen mode //working but issues with windows resizing on toggle
		# // int $inFullScreenMode = `optionVar -q "workspacesInFullScreenUIMode"`;
		# // int $inZoomInMode = `optionVar -q "workspacesInZoomInUIMode"`;
		# // // enter full screen mode only if the zoom-in mode is not active.
		# // if(!$inZoomInMode) 
		# // {
		# // 	string $panelWithFocus = `getPanel -withFocus`;
		# // 	string $parentControl = `workspaceLayoutManager -parentWorkspaceControl $panelWithFocus`;
		# // 	int $isFloatingPanel = `workspaceControl -q -floating $parentControl`;
						
		# // 	if(!$isFloatingPanel) 
		# // 	{
		# // 		if($inFullScreenMode) 
		# // 		{
		# // 		//come out of fullscreen mode
		# // 		workspaceLayoutManager -restoreMainWindowControls;
		# // 		}
		# // 		else 
		# // 		{
		# // 			// enter fullscreen mode
		# // 			workspaceLayoutManager -collapseMainWindowControls $parentControl true;
		# // 		}
		# // 	optionVar -iv "workspacesInFullScreenUIMode" (!$inFullScreenMode);
		# // 	}
		# // }


import os
import setuptools

from mayatk import name, __version__
from pythontk import File


with open('docs/README.md', 'r') as f:
	long_description = f.read()

setuptools.setup(
	name=name,
	version=__version__,
	author='Ryan Simpson',
	author_email='m3trik@outlook.com',
	description='.',
	long_description=long_description,
	long_description_content_type='text/markdown',
	url=f'https://github.com/m3trik/{name}',
	packages=setuptools.find_packages(exclude=['pymel', 'pymel.*']),
	classifiers=[
		'Programming Language :: Python :: 3',
		'License :: OSI Approved :: MIT License',
		'Operating System :: OS Independent',
	],
	include_package_data = True,
	data_files=File.getDirContents(name, 'filepaths', excFiles=['*.py', '*.pyc', '*.json']), #ie. ('tentacle/ui/0', ['tentacle/ui/0/init.ui']),
)

# --------------------------------------------------------------------------------------------









# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# Deprecated ------------------------------------


# def gen_data_files(dirs, exc=[], inc=[]):
# 	'''
# 	'''
# 	dirs = itertk.makeList(dirs)
# 	exc = itertk.makeList(exc)
# 	inc = itertk.makeList(inc)

# 	results = []
# 	for src_dir in dirs:
# 		for root, dirs, files in os.walk(src_dir):
# 			filtered=[]
# 			for f in files:
# 				ext = File.formatPath(f, 'ext')
# 				if f in exc or '*.'+ext in exc:
# 					continue
# 				if any(inc): #filter inc for None values so not to get a false positive.
# 					if not f in inc and not '*.'+ext in inc:
# 						continue
# 				filtered.append(f)

# 			if filtered:
# 				results.append((root, list(map(lambda f:root + "/" + f, filtered))))
# 	return results

# # for i in gen_data_files('tentacle', exc=['*.py', '*.pyc', '*.json']):
# 	# print (i)


# !/usr/bin/python
# coding=utf-8
import os

try:
	import pymel.core as pm
	import maya.OpenMaya as om
	import maya.OpenMayaFX as omfx

	import MASH.api as mapi

except ImportError as error:
	print (__file__, error)


def MTcreateNetwork(network, objects, networkName='MASH#', geometry='Mesh', distType='linearNetwork', hideOnCreate=True):
	'''This proceedure creates a new MASH network.

	:Parameters:
		network (obj) = A mash api network instance. ie. mapi.Network()
		objects (str)(obj)(list) = The maya objects to construct the network for.
		networkName (str) = The network name. (default:'MASH#')
		geometry (str) = Particle instancer or mesh instancer (Repro node). (valid: 'Mesh' (default), 'Instancer')
		distType (str) = Distribution type. (valid: "linearNetwork", radialNetwork", "gridNetwork", "initialNetwork", "zeroNetwork")
		hideOnCreate (bool) = Hide the original object(s) on network creation.

	:Return:
		The names of the created Waiter, Instancer and Distribute Node.

	ex. mashNW = mapi.Network() #network instance
		mashNW.MTcreateNetwork(objects, geometry='Instancer', hideOnCreate=False) #create a MASH network
	'''
	loaded = int(pm.pluginInfo('MASH', query=1, l=1))
	if not loaded:
		pm.loadPlugin('MASH')

	waiter=distNode=instancer = ''

	if geometry=='Mesh':
		objects_ = pm.ls(obj, lf=1, ni=1, dag=1, type="mesh", l=1) #filter for valid mesh objects.
	else:
		objects_ = pm.ls(objects)

	if len(objects_)==0:
		return pm.mel.MASHinViewMessage((pm.mel.getPluginResource("MASH", "kMASHMeshesOnly")), "Error") #error and bail if there was nothing usable selected

	if hideOnCreate:
		[pm.hide(obj) for obj in objects_] #hide the original objects

	waiter = pm.ls(pm.createNode('MASH_Waiter', n=networkName))[0] #create the waiter
	pm.addAttr(waiter, hidden=True, at='message', longName='instancerMessage')

	distNodeName = '{}_Distribute'.format(waiter.name())
	distNode = pm.ls(pm.createNode('MASH_Distribute', n=distNodeName))[0] #create a Distribute node

	pm.setAttr(distNode.mapDirection, 4) #for convenience we match the Map Direction to the initial linear distribution

	#connect the Distribute node to the Waiter
	pm.connectAttr(distNode.outputPoints, waiter.inputPoints, force=1)
	pm.connectAttr(distNode.waiterMessage, waiter.waiterMessage, f=1)

	if geometry=='Mesh': #create the repro node
		import mash_repro_utils; reload(mash_repro_utils)
		reproName = '{}_Repro'.format(waiter.name())		
		instancer = pm.ls(mash_repro_utils.create_mash_repro_node(None, reproName))[0]
	else: # create the instancer
		instancerName = '{}_Instancer'.format(waiter.name())
		instancer = pm.ls(pm.createNode('instancer', name=instancerName))[0]

	pm.connectAttr(waiter.outputPoints, instancer.inputPoints, force=1) #connect the Waiter to the Instancer or Repro node
	pm.addAttr(instancer, hidden=True, at='message', longName='instancerMessage') # add a message attribute to the instancer
	pm.connectAttr(waiter.instancerMessage, instancer.instancerMessage, f=1) #connect message attributes to the instancer so we can find it later if needed

	#if the user added more then one object to the network, set the number of points to reflect that
	size = len(objects_)
	if len(objects_)>1:
		pm.setAttr(distNode.pointCount, size)

	# add the objects to the instancer or Repro
	for transform in pm.ls(objects_, transforms=1):
		if geometry=='Mesh':
			import mash_repro_utils
			mash_repro_utils.connect_mesh_group(instancer.name(), transform.name(), new_connection=True) # connect the meshes to the Repro node
		else:
			import maya.mel as mel
			pm.mel.eval('instancer -e -a -obj {} {};'.format(transform, instancer.name())) #pm.instancer(instancerName, e=1, addObject=1, obj=transform)

	#set the Distribute node defaults
	if distType=="radialNetwork":
		pm.setAttr(distNode.arrangement, 2)
	elif distType=="gridNetwork":
		pm.setAttr(distNode.arrangement, 6)
	elif distType=="initialNetwork":
		pm.setAttr(distNode.arrangement, 7)
	elif distType=="zeroNetwork":
		pm.setAttr(distNode.amplitudeX, 0.0)

	#update the Repro interface
	if geometry=='Mesh':
		import mash_repro_aetemplate
		mash_repro_aetemplate.refresh_all_aetemplates(force=True)

	coreNodes = {'waiter':waiter, 'instancer':instancer, 'distribute':distNode}
	#add created nodes to the network.
	for k,v in coreNodes.items():
		if not hasattr(network, k):
			setattr(network, k, v)


	return coreNodes.values()


def MTbakeInstancer(network, instancer, bakeTranslate=True, bakeRotation=True, bakeScale=True, 
	bakeAnimation=False, bakeVisibility=True, bakeToIntances=False, upstreamNodes=False, _getMObjectFromName=None):
	'''Takes an instancer, and coverts all the particles being fed into it to actual geometry.

	:Parameters:
		network (obj) = A mash api network instance. ie. mapi.Network()
		instancer (str)(obj) = Instancer node or name of an instancer node.
		bakeAnimation (bool) = True=bake entire playback range, False=bake current frame.
		bakeTranslate (bool) = Bake translation.
		bakeRotation (bool) = Bake rotation.
		bakeScale (bool) = Bake scale.
		bakeVisibility (bool) = Bake visibility.
		bakeToInstances (bool) = Bake to instances rather than separate objects.
		upstreamNodes (bool) = The upstream nodes leading upto the selected nodes (along with their connections) are also duplicated.
		_getMObjectFromName (str) = Node name. Returns a maya.OpenMaya.MObject from the given instancer node name. Internal use only.

	:Return:
		(list) Baked objects.
	'''
	if _getMObjectFromName:
		sel = om.MSelectionList()
		sel.add(_getMObjectFromName)
		thisNode = om.MObject()
		sel.getDependNode(0, thisNode)
		return thisNode

	_instancer = pm.ls(instancer, type='instancer')[0]
	if not _instancer: #check that the nodeType is 'instancer'.
		raise Exception('"{}" is type: "{}". The required node type is "instancer".'.format(instancer, pm.nodeType(instancer)))

	#reused vars for the particles
	m = om.MMatrix()
	dp = om.MDagPath()
	dpa = om.MDagPathArray()
	sa = om.MScriptUtil()
	sa.createFromList([0.0, 0.0, 0.0], 3)
	sp = sa.asDoublePtr()

	#Get the instancer function set
	thisNode = MTbakeInstancer(network, _instancer, _getMObjectFromName=_instancer.name())
	fnThisNode = om.MFnDependencyNode(thisNode)

	#start frame, end frame, animaiton
	sf = int(pm.playbackOptions(q=True, min=True))-1
	ef = int(pm.playbackOptions(q=True, max=True))+2

	if bakeAnimation==False:
		sf = pm.currentTime(query=True)
		ef = sf+1

	result=[]
	for i in range(int(sf), int(ef)):
		#set the time
		pm.currentTime(i)
		g = '{}_objects'.format(_instancer.name())
		
		#get the visibility array - which isn't provided by the MFnInstancer function set
		inPointsAttribute = fnThisNode.attribute("inputPoints")
		inPointsPlug = om.MPlug( thisNode, inPointsAttribute )
		inPointsObj = inPointsPlug.asMObject()
		inputPPData = om.MFnArrayAttrsData(inPointsObj)
		visList = inputPPData.getDoubleData("visibility")[:]

		#if this is the first frame, create a transform to store everything under
		if i==sf:
			if pm.objExists(g)==True:
				pm.delete(g)
			g = pm.createNode("transform", n=g)

		#get the instancer
		sl = om.MSelectionList()
		sl.add(_instancer)
		sl.getDagPath(0, dp)
		#create mfninstancer function set
		fni = omfx.MFnInstancer(dp)

		#cycle through the particles
		for j in range(fni.particleCount()):
			visibility = visList[j]
			#get the instancer object
			fni.instancesForParticle(j, dpa, m)
			for i in range(dpa.length()):
				fullPathName = dpa[i].partialPathName() #get the instancer object name
				#support namespaces, refrences, crap names
				nameSpaceRemoved = fullPathName.rsplit(':', 1)[-1]
				pipesRemoved = nameSpaceRemoved.rsplit('|', 1)[-1]
				name = '{}_{}_{}'.format(instancer.name(), pipesRemoved, j)

				#duplicate the object
				if bakeToIntances:
					name = pm.instance(dpa[i].fullPathName(), leaf=1, name=name)[0]
				else:
					name = pm.duplicate(dpa[i].fullPathName(), returnRootsOnly=1, upstreamNodes=upstreamNodes, name=name)[0]

				#parent it to the transform we created above
				if pm.listRelatives(name, p=True) != g:
					try:
						name = pm.parent(name, g)[0]
					except:
						pass

					# if the object doesn't appear on frame 0(animated creation), set the visibility when it first appears
					pm.setKeyframe(name+".visibility", v=0, t=pm.currentTime(query=1)-1)
					pm.setKeyframe(name+".visibility", v=1)

				#empty transformMatrix for the particle
				tm = om.MTransformationMatrix(m)
				instancedPath = dpa[i]
				#get the matrix from the instancer
				instancedPathMatrix = instancedPath.inclusiveMatrix()
				finalMatrixForPath = instancedPathMatrix * m
				finalPoint = om.MPoint.origin * finalMatrixForPath;

				t = tm.getTranslation(om.MSpace.kWorld)
				#set the translate
				try:
					pm.setAttr(name+".t", finalPoint.x, finalPoint.y, finalPoint.z)
					if bakeTranslate and bakeAnimation:
						pm.setKeyframe(name+".t")
				except:
					pass

				#set the rotate
				r = tm.eulerRotation().asVector()
				try:
					pm.setAttr(name+".r", r[0]*57.2957795, r[1]*57.2957795, r[2]*57.2957795)
					if bakeRotation and bakeAnimation:
						pm.setKeyframe(name+".r")
				except:
					pass

				#set the scale
				tm.getScale(sp, om.MSpace.kWorld)
				if bakeScale:
					sx = om.MScriptUtil().getDoubleArrayItem(sp, 0)
					sy = om.MScriptUtil().getDoubleArrayItem(sp, 1)
					sz = om.MScriptUtil().getDoubleArrayItem(sp, 2)
					s = om.MTransformationMatrix(dpa[i].inclusiveMatrix()).getScale(sp, om.MSpace.kWorld)
					sx2 = om.MScriptUtil().getDoubleArrayItem(sp, 0)
					sy2 = om.MScriptUtil().getDoubleArrayItem(sp, 1)
					sz2 = om.MScriptUtil().getDoubleArrayItem(sp, 2)
					try:
						pm.setAttr(name+".s", sx*sx2, sy*sy2, sz*sz2)
						if bakeAnimation:
							pm.setKeyframe(name+".s")
					except:
						pass

				# set visibility
				if bakeVisibility:
					pm.setAttr(name+".v", visibility)
					if bakeAnimation:
						pm.setKeyframe(name+".v")

				result.append(name) #append the converted object to the return value.

		return result


try:
	# add each funtion as an network attribute.
	for k, v in {
		'MTcreateNetwork':MTcreateNetwork, 
		'MTbakeInstancer':MTbakeInstancer,
		}.items():

		if not hasattr(mapi.Network, k):
			setattr(mapi.Network, k, v)

except NameError as error: #import mapi failed.
	print (__file__, error)

# --------------------------------------------------------------------------------------------









# print (__name__) #module name
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# deprecated: -----------------------------------



# !/usr/bin/python
# coding=utf-8
import sys, os

from PySide2 import QtCore, QtWidgets

import pymel.core as pm

from pythontk import File, Img, Str, Json
from mayatk import Node, getMainWindow
from tentacle.switchboard import Switchboard
from tentacle.widgets import rwidgets

__version__ = '0.5.3'

class Stingray_arnold_shader(QtCore.QObject):
	'''
	To correctly render opacity and transmission, the Opaque setting needs to be disabled on the Shape node.
	If Opaque is enabled, opacity will not work at all. Transmission will work however any shadows cast by 
	the object will always be solid and not pick up the Transparent Color or density of the shader.
	'''
	msg_intro = '''<u>To setup the material:</u>
		<br>• Use the <b>Get Texture Maps</b> button to load the images you intend to use,
		<br>and click the <b>Create Network</b> button to create the shader connections.
		<br>• The HDR map, it's visiblity, and rotation can be changed after creation.
	'''
	msg_completed = '<br><hl style="color:rgb(0, 255, 255);"><b>COMPLETED.</b></hl>'

	proj_root_dir = File.getFilepath(__file__)
	hdr_env_name = 'aiSkyDomeLight_'

	def __init__(self, parent=None):
		super().__init__(parent)


	@property
	def hdr_env(self) -> object:
		'''
		'''
		node = pm.ls(self.hdr_env_name, exactType='aiSkyDomeLight')
		try:
			return node[0]
		except IndexError as error:
			return None


	@hdr_env.setter
	def hdr_env(self, tex) -> None:
		'''
		'''
		node = self.hdr_env #Node.nodeExists('aiSkyDomeLight', search='exactType')
		if not node:
			node = Node.createRenderNode('aiSkyDomeLight', 'asLight', name=self.hdr_env_name, camera=0, skyRadius=0) #turn off skydome and viewport visibility.
			self.hdr_env_transform.hiddenInOutliner.set(1)
			pm.outlinerEditor('outlinerPanel1', edit=True, refresh=True)

		file_node = Node.getIncomingNodeByType(node, 'file')
		if not file_node:
			file_node = Node.createRenderNode('file', 'as2DTexture', place2dTexture=True)
			pm.connectAttr(file_node.outColor, node.color, force=True)

		file_node.fileTextureName.set(tex)


	@property
	def hdr_env_transform(self) -> object:
		'''
		'''
		node = Node.getTransformNode(self.hdr_env)
		if not node:
			return None
		return node


	def setHdrMapVisibility(self, state):
		'''
		'''
		node = self.hdr_env
		if node:
			node.camera.set(state)


	def createNetwork(self, textures, name='', hdrMap='', hdrMapVisibility=False, normalMapType='OpenGL', callback=print):
		'''
		'''
		pm.undoChunk(open=True)

		normal_map_created_from_other_type = False
		normalMapType = 'Normal_'+normalMapType.strip('Normal_') #assure normalMapType is formatted as 'Normal_OpenGL' whether given as 'OpenGL' or 'Normal_OpenGL'

		if not textures:
			callback('<br><hl style="color:rgb(255, 100, 100);"><b>Error:</b> No textures given to createNetwork.</hl>')
			return None
		try:
			pm.loadPlugin('mtoa', quiet=True) #assure arnold plugin is loaded.
			pm.loadPlugin('shaderFXPlugin', quiet=True) #assure stringray plugin is loaded.

			sr_node = Node.createRenderNode('StingrayPBS', name=name)
			ai_node = Node.createRenderNode('aiStandardSurface', name=name+'_ai' if name else '')

			opacityMap = Img.filterImagesByType(textures, 'Opacity')
			if opacityMap:
				pm.shaderfx(sfxnode='StingrayPBS1', loadGraph=r'C:/_local/_test/shaderfx/Standard_Transparent.sfx')

			openGLMap = Img.filterImagesByType(textures, 'Normal_OpenGL')
			directXMap = Img.filterImagesByType(textures, 'Normal_DirectX')
			if directXMap and not openGLMap and normalMapType=='Normal_OpenGL':
				mapPath = Img.createGLFromDX(directXMap[0])
				textures.append(mapPath)
				normal_map_created_from_other_type = True
				callback('OpenGL map created using {}.'.format(Str.truncate(directXMap[0], 20)))
			if openGLMap and not directXMap and normalMapType=='Normal_DirectX':
				mapPath = Img.createDXFromGL(openGLMap[0])
				textures.append(mapPath)
				normal_map_created_from_other_type = True
				callback('DirectX map created using {}.'.format(Str.truncate(openGLMap[0], 20)))

			srSG_node = Node.getOutgoingNodeByType(sr_node, 'shadingEngine')

			aiMult_node = pm.shadingNode('aiMultiply', asShader=True)

			bump_node = pm.shadingNode('bump2d', asShader=True)
			bump_node.bumpInterp.set(1) #set bump node to 'tangent space normals'

			Node.connectMultiAttr( #set node connections.
				(ai_node.outColor, srSG_node.aiSurfaceShader),
				(aiMult_node.outColor, ai_node.baseColor),
				(bump_node.outNormal, ai_node.normalCamera),
			)

			length = len(textures)
			progress = 0
			for f in textures:
				typ = Img.getImageTypeFromFilename(f)

				progress+=1

				#filter normal maps for the correct type.
				if typ=='Normal' and (openGLMap or directXMap):
					continue
				elif typ=='Normal_DirectX' and normalMapType=='Normal_OpenGL':
					continue
				elif typ=='Normal_OpenGL' and normalMapType=='Normal_DirectX':
					continue

				callback('creating nodes and connections for <b>{}</b> map ..'.format(typ), [progress, length])

				if typ=='Base_Color':
					n1 = Node.createRenderNode('file', 'as2DTexture', tex=f)
					pm.connectAttr(n1.outColor, sr_node.TEX_color_map, force=True)
					sr_node.use_color_map.set(1)

					n2 = Node.createRenderNode('file', 'as2DTexture', tex=f, place2dTexture=True)
					pm.connectAttr(n2.outColor, aiMult_node.input1, force=True)

				elif typ=='Roughness':
					n1 = Node.createRenderNode('file', 'as2DTexture', tex=f)
					pm.connectAttr(n1.outColor, sr_node.TEX_roughness_map, force=True)
					sr_node.use_roughness_map.set(1)

					n2 = Node.createRenderNode('file', 'as2DTexture', tex=f, place2dTexture=True, colorSpace='Raw', alphaIsLuminance=1, ignoreColorSpaceFileRules=1)
					pm.connectAttr(n2.outAlpha, ai_node.specularRoughness, force=True)
					pm.connectAttr(n2.outAlpha, ai_node.transmissionExtraRoughness, force=True) #opacity: same roughness map used in Specular Roughness to provide additional bluriness of refraction.

				elif typ=='Metallic':
					n1 = Node.createRenderNode('file', 'as2DTexture', tex=f)
					pm.connectAttr(n1.outColor, sr_node.TEX_metallic_map, force=True)
					sr_node.use_metallic_map.set(1)

					n2 = Node.createRenderNode('file', 'as2DTexture', tex=f, place2dTexture=True, colorSpace='Raw', alphaIsLuminance=1, ignoreColorSpaceFileRules=1)
					pm.connectAttr(n2.outAlpha, ai_node.metalness, force=True)

				elif typ=='Emissive':
					n1 = Node.createRenderNode('file', 'as2DTexture', tex=f)
					pm.connectAttr(n1.outColor, sr_node.TEX_emissive_map, force=True)
					sr_node.use_emissive_map.set(1)

					n2 = Node.createRenderNode('file', 'as2DTexture', tex=f, place2dTexture=True)
					pm.connectAttr(n2.outAlpha, ai_node.emission, force=True)
					pm.connectAttr(n2.outColor, ai_node.emissionColor, force=True)

				elif 'Normal' in typ:
					n1 = Node.createRenderNode('file', 'as2DTexture', tex=f)
					pm.connectAttr(n1.outColor, sr_node.TEX_normal_map, force=True)
					sr_node.use_normal_map.set(1)

					n2 = Node.createRenderNode('file', 'as2DTexture', tex=f, place2dTexture=True, colorSpace='Raw', alphaIsLuminance=1, ignoreColorSpaceFileRules=1)
					pm.connectAttr(n2.outAlpha, bump_node.bumpValue, force=True)

				elif typ=='Ambient_Occlusion':
					n1 = Node.createRenderNode('file', 'as2DTexture', tex=f)
					pm.connectAttr(n1.outColor, sr_node.TEX_ao_map, force=True)
					sr_node.use_ao_map.set(1)

					n2 = Node.createRenderNode('file', 'as2DTexture', tex=f, place2dTexture=True)
					pm.connectAttr(n2.outColor, aiMult_node.input2, force=True)

				elif typ=='Opacity':
					n1 = Node.createRenderNode('file', 'as2DTexture', tex=f)
					pm.connectAttr(n1.outAlpha, sr_node.opacity, force=True)
					sr_node.use_opacity_map.set(1)

					n2 = Node.createRenderNode('file', 'as2DTexture', tex=f, place2dTexture=True, colorSpace='Raw', alphaIsLuminance=1, ignoreColorSpaceFileRules=1)
					pm.connectAttr(n2.outAlpha, ai_node.transmission, force=True)
					pm.connectAttr(n2.outColor, ai_node.opacity, force=True)

				else:
					if normal_map_created_from_other_type:
						continue #do not show a warning for unconnected normal maps if it resulted from being converted to a different output type.
					callback('<br><hl style="color:rgb(255, 100, 100);"><b>Map type: <b>{}</b> not connected:<br></hl>'.format(typ, Str.truncate(f, 60)), [progress, length])
					continue

				callback('<font style="color: rgb(80,180,100)">{}..connected successfully.</font>'.format(typ))
		except Exception as error:
			callback('<br><hl style="color:rgb(255, 100, 100);"><b>Error:</b>{}.</hl>'.format(error))

		self.hdr_env = hdrMap
		self.setHdrMapVisibility(hdrMapVisibility)
		pm.undoChunk(close=True)



class Stingray_arnold_shader_slots(Stingray_arnold_shader):

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		'''
		'''
		self.ui = self.sb.stingray_arnold_shader #slot classes used with the switchboard module inherit the 'sb' attribute by default.

		self.imageFiles = None

		#set json file location.
		path = '{}/stingray_arnold_shader.json'.format(self.sb.defaultDir)
		Json.setJsonFile(path) #set json file name

		#add filenames|filepaths to the comboBox.
		hdr_path = '{}/resources/hdr'.format(self.proj_root_dir)
		hdr_filenames = File.getDirContents(hdr_path, 'filenames', incFiles='*.exr')
		hdr_fullpaths = File.getDirContents(hdr_path, 'filepaths', incFiles='*.exr')
		self.ui.cmb000.addItems_(dict(zip(hdr_filenames, hdr_fullpaths)), ascending=False)

		#initialize widgets with any saved values.
		self.ui.txt000.setText(Json.getJson('mat_name'))
		self.ui.txt001.setText(self.msg_intro)
		hdr_map_visibility = Json.getJson('hdr_map_visibility')
		if hdr_map_visibility:
			self.ui.chk000.setChecked(hdr_map_visibility)
		hdr_map = Json.getJson('hdr_map')
		if hdr_map:
			self.ui.cmb000.setCurrentItem(hdr_map)
		normal_map_type = Json.getJson('normal_map_type')
		if normal_map_type:
			self.ui.cmb001.setCurrentItem(normal_map_type)
		node = self.hdr_env_transform
		if node:
			rotation = node.rotateY.get()
			self.ui.slider000.setSliderPosition(rotation)


	@property
	def mat_name(self) -> str:
		'''Get the mat name from the user input text field.

		:Return:
			(str)
		'''
		text = self.ui.txt000.text()
		return text


	@property
	def hdr_map(self) -> str:
		'''Get the hdr map filepath from the comboBoxes current text.

		:Return:
			(str) data as string.
		'''
		data = self.ui.cmb000.currentData()
		return data


	@property
	def hdr_map_visibility(self) -> bool:
		'''Get the hdr map visibility state from the checkBoxes current state.

		:Return:
			(bool)
		'''
		state = self.ui.chk000.isChecked()
		return state


	@property
	def normal_map_type(self) -> str:
		'''Get the normal map type from the comboBoxes current text.

		:Return:
			(str)
		'''
		text = self.ui.cmb001.currentText()
		return text


	def cmb000(self, index):
		'''HDR map selection.
		'''
		cmb = self.ui.cmb000
		text = cmb.currentText()
		data = cmb.currentData()

		self.hdr_env = data #set the HDR map.
		Json.setJson('hdr_map', text)


	def chk000(self, state):
		'''
		'''
		chk = self.ui.chk000

		self.setHdrMapVisibility(state) #set the HDR map visibility.
		Json.setJson('hdr_map_visibility', state)


	def cmb001(self, index):
		'''Normal map output selection.
		'''
		cmb = self.ui.cmb001
		text = cmb.currentText()
		Json.setJson('normal_map_type', text)


	def txt000(self, text=None):
		'''Material name.
		'''
		txt = self.ui.txt000
		text = txt.text()
		Json.setJson('mat_name', text)


	def slider000(self, value):
		'''Rotate the HDR map.
		'''
		if self.hdr_env:
			transform = Node.getTransformNode(self.hdr_env)
			pm.rotate(transform, value, rotateY=True, forceOrderXYZ=True, objectSpace=True, absolute=True)


	def b000(self):
		'''Create network.
		'''
		if self.imageFiles:
			# pm.mel.HypershadeWindow() #open the hypershade window.

			self.ui.txt001.clear()
			self.callback('Creating network ..<br>')

			self.createNetwork(self.imageFiles, self.mat_name, hdrMap=self.hdr_map,
				hdrMapVisibility=self.hdr_map_visibility, normalMapType=self.normal_map_type, 
				callback=self.callback)

			self.callback(self.msg_completed)
			# pm.mel.hyperShadePanelGraphCommand('hyperShadePanel1', 'rearrangeGraph')


	def b001(self):
		'''Get texture maps.
		'''
		imageFiles = Img.getImageFiles()
		if imageFiles:
			self.imageFiles = imageFiles
			self.ui.txt001.clear()

			msg_mat_selection = self.imageFiles
			for i in msg_mat_selection: #format msg_intro using the mapTypes in imtools.
				self.callback(Str.truncate(i, 60))

			self.ui.b000.setDisabled(False)
		elif not self.imageFiles:
			self.ui.b000.setDisabled(True)


	def callback(self, string, progress=None, clear=False):
		'''
		:Parameters:
			string (str) = The text to output to a textEdit widget.
			progress (int)(list) = The progress amount to register with the progressBar.
				Can be given as an int or a tuple as: (progress, total_len) 
		'''
		if clear:
			self.ui.txt003.clear()

		if isinstance(progress, (list, tuple, set)):
			p, l = progress
			progress = (p/l) *100

		self.ui.txt001.append(string)

		if progress is not None:
			self.ui.progressBar.setValue(progress)
			self.sb.app.processEvents()


class Stingray_arnold_shader_main(Stingray_arnold_shader):
	'''
	'''
	app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv) #return the existing QApplication object, or create a new one if none exists.

	def __init__(self, parent=None):
		super().__init__(parent)

		self.sb = Switchboard(self, uiLoc='stingray_arnold_shader.ui', widgetLoc=rwidgets, slotLoc=Stingray_arnold_shader_slots)

		ui = self.sb.stingray_arnold_shader
		print('slots:', self.sb.getSlots(ui))
		self.sb.setStyle(ui.widgets)
		ui.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)
		ui.show()

# -----------------------------------------------------------------------------









# -----------------------------------------------------------------------------

if __name__ == "__main__":

	parent = getMainWindow()
	main = Stingray_arnold_shader_main(parent)

	# sys.exit(main.app.exec_()) # run app, show window, wait for input, then terminate program with a status code returned from app.

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------



# Deprecated ------------------------------------------------------------------


# !/usr/bin/python
# coding=utf-8
try:
	import pymel.core as pm
except ImportError as error:
	print (__file__, error)

from pythontk import Iter


class Node():
	'''
	'''
	@staticmethod
	def nodeExists(n, search='name'):
		'''Check if the node exists in the current scene.

		:Parameters:
			search (str) = The search parameters. valid: 'name', 'type', 'exactType'

		:Return:
			(bool)
		'''
		if search=='name':
			return bool(pm.ls(n))
		elif search=='type':
			return bool(pm.ls(type=n))
		elif search=='exactType':
			return bool(pm.ls(exactType=n))


	@classmethod
	def getType(cls, objects):
		'''Get the object type as a string.

		:Parameters:
			objects (str)(obj)(list) = The object(s) to query.

		:Return:
			(str)(list) The node type. A list is always returned when 'objects' is given as a list.
		'''
		types=[]
		for obj in pm.ls(objects):

			if cls.isGroup(obj):
				typ = 'group'
			elif cls.isLocator(obj):
				typ = 'locator'
			else:
				typ = comptk.getComponentType(obj)
			if not typ:
				typ = pm.objectType(obj)
			types.append(typ)

		return Iter.formatReturn(types, objects)


	@classmethod
	def isLocator(cls, obj):
		'''Check if the object is a locator.
		A locator is a transform node that has a shape node child.
		The shape node defines the appearance and behavior of the locator.

		:Parameters:
			obj () = The object to query.

		:Return:
			(bool)
		'''
		shape = cls.getShapeNode(obj)
		return pm.nodeType(shape)=='locator'


	@staticmethod
	def isGroup(objects):
		'''Determine if each of the given object(s) is a group.
		A group is defined as a transform with children.

		:Parameters:
			nodes (str)(obj)(list) = The object(s) to query.

		:Return:
			(bool)(list) A list is always returned when 'objects' is given as a list.
		'''
		result=[]
		for n in pm.ls(objects):
			try:
				q = all((
					type(n)==pm.nodetypes.Transform,
					all(([type(c)==pm.nodetypes.Transform for c in n.getChildren()])),
				))
			except AttributeError as error:
				q = False
			result.append(q)

		return Iter.formatReturn(result, objects)


	@classmethod
	def getGroups(cls, empty=False):
		'''Get all groups in the scene.

		:Parameters:
			empty (bool) = Return only empty groups.

		:Return:
			(bool)
		'''
		transforms =  pm.ls(type='transform')

		groups=[]
		for t in transforms:
			if cls.isGroup(t):
				if empty:
					children = pm.listRelatives(t, children=True)
					if children:
						continue
				groups.append(t)

		return groups


	@staticmethod
	def getParent(node, all=False):
		'''List the parents of an object.
		'''
		if all:
			objects = pm.ls(node, l=1)
			tokens=[]
			return objects[0].split("|")

		try:
			return pm.listRelatives(node, parent=True, type='transform')[0]
		except IndexError as error:
			return None


	@staticmethod
	def getChildren(node):
		'''List the children of an object.
		'''
		try:
			return  pm.listRelatives(node, children=True, type='transform')
		except IndexError as error:
			return []


	@staticmethod
	def getTransformNode(nodes, returnType='obj', attributes=False, inc=[], exc=[]):
		'''Get transform node(s) or node attributes.

		:Parameters:
			nodes (str)(obj)(list) = A relative of a transform Node.
			returnType (str) = The desired returned object type. Not valid with the `attributes` parameter.
				(valid: 'str'(default), 'obj').
			attributes (bool) = Return the attributes of the node, rather then the node itself.

		:Return:
			(obj)(list) node(s) or node attributes. A list is always returned when 'nodes' is given as a list.
		'''
		from mayatk import convertArrayType

		result=[]
		for node in pm.ls(nodes):
			transforms = pm.ls(node, type='transform')
			if not transforms: #from shape
				shapeNodes = pm.ls(node, objectsOnly=1)
				transforms = pm.listRelatives(shapeNodes, parent=1)
				if not transforms: #from history
					try:
						transforms = pm.listRelatives(pm.listHistory(node, future=1), parent=1)
					except Exception as error:
						transforms=[]
			for n in transforms:
				result.append(n)

		if attributes:
			result = pm.listAttr(result, read=1, hasData=1)

		#convert element type.
		result = convertArrayType(result, returnType=returnType, flatten=True)
		#filter
		result = Iter.filterList(result, inc, exc)
		#return as list if `nodes` was given as a list.
		return Iter.formatReturn(list(set(result)), nodes)


	@classmethod
	def getShapeNode(cls, nodes, returnType='obj', attributes=False, inc=[], exc=[]):
		'''Get shape node(s) or node attributes.

		:Parameters:
			nodes (str)(obj)(list) = A relative of a shape Node.
			returnType (str) = The desired returned object type. 
				(valid: 'str'(default), 'obj'(shape node), 'transform'(as string), 'int'(valid only at sub-object level).
			attributes (bool) = Return the attributes of the node, rather then the node itself.

		:Return:
			(obj)(list) node(s) or node attributes. A list is always returned when 'nodes' is given as a list.
		'''
		from mayatk import convertArrayType

		result=[]
		for node in pm.ls(nodes):
			shapes = pm.listRelatives(node, children=1, shapes=1) #get shape node from transform: returns list ie. [nt.Mesh('pConeShape1')]
			if not shapes:
				shapes = pm.ls(node, type='shape')
				if not shapes: #get shape from transform
					try:
						transforms = pm.listRelatives(pm.listHistory(node, future=1), parent=1)
						shapes = cls.getShapeNode(transforms)
					except Exception as error:
						shapes = []
			for n in shapes:
				result.append(n)

		if attributes:
			result = pm.listAttr(result, read=1, hasData=1)

		#convert element type.
		result = convertArrayType(result, returnType=returnType, flatten=True)
		#filter
		result = Iter.filterList(result, inc, exc)
		#return as list if `nodes` was given as a list.
		return Iter.formatReturn(list(set(result)), nodes)


	@staticmethod
	def getHistoryNode(nodes, returnType='obj', attributes=False, inc=[], exc=[]):
		'''Get history node(s) or node attributes.

		:Parameters:
			nodes (str)(obj)(list) = A relative of a history Node.
			returnType (str) = The desired returned object type. 
				(valid: 'str'(default), 'obj'(shape node), 'transform'(as string), 'int'(valid only at sub-object level).
			attributes (bool) = Return the attributes of the node, rather then the node itself.

		:Return:
			(obj)(list) node(s) or node attributes. A list is always returned when 'nodes' is given as a list.
		'''
		from mayatk import convertArrayType

		result=[]
		for node in pm.ls(nodes):
			shapes = pm.listRelatives(node, children=1, shapes=1) #get shape node from transform: returns list ie. [nt.Mesh('pConeShape1')]
			try:
				history = pm.listConnections(shapes, source=1, destination=0)[-1] #get incoming connections: returns list ie. [nt.PolyCone('polyCone1')]
			except IndexError as error:
				try:
					history = node.history()[-1]
				except AttributeError as error:
					print ('{} in getHistoryNode\n\t# Error: {} #'.format(__file__, error))
					continue
			result.append(history)

		if attributes:
			result = pm.listAttr(result, read=1, hasData=1)

		#convert element type.
		result = convertArrayType(result, returnType=returnType, flatten=True)
		#filter
		result = Iter.filterList(result, inc, exc)
		#return as list if `nodes` was given as a list.
		return Iter.formatReturn(list(set(result)), nodes)


	@classmethod
	def createRenderNode(cls, nodeType, flag='asShader', flag2='surfaceShader', name='', tex='', place2dTexture=False, postCommand='', **kwargs):
		'''Procedure to create the node classified as specified by the inputs.

		:Parameters:
			nodeType (str) = The type of node to be created. ie. 'StingrayPBS' or 'aiStandardSurface'
			flag (str) = A flag specifying which how to classify the node created.
				valid:	as2DTexture, as3DTexture, asEnvTexture, asShader, asLight, asUtility
			flag2 (str) = A secondary flag used to make decisions in combination with 'asType'
				valid:	-asBump : defines a created texture as a bump
						-asNoShadingGroup : for materials; create without a shading group
						-asDisplacement : for anything; map the created node to a displacement material.
						-asUtility : for anything; do whatever the $as flag says, but also classify as a utility
						-asPostProcess : for any postprocess node
			name (str) = The desired node name.
			tex (str) = The path to a texture file for those nodes that support one.
			place2dTexture (bool) = If not needed, the place2dTexture node will be deleted after creation.
			postCommand (str) = A command entered by the user when invoking createRenderNode.
					The command will substitute the string %node with the name of the
					node it creates.  createRenderWindow will be closed if a command
					is not the null string ("").
			kwargs () = Set additional node attributes after creation. ie. colorSpace='Raw', alphaIsLuminance=1, ignoreColorSpaceFileRules=1

		:Return:
			(obj) node

		ex. call: createRenderNode('StingrayPBS')
		ex. call: createRenderNode('file', flag='as2DTexture', tex=f, place2dTexture=True, colorSpace='Raw', alphaIsLuminance=1, ignoreColorSpaceFileRules=1)
		ex. call: createRenderNode('aiSkyDomeLight', tex=pathToHdrMap, name='env', camera=0, skyRadius=0) #turn off skydome and viewport visibility.
		'''
		node = pm.PyNode(pm.mel.createRenderNodeCB('-'+flag, flag2, nodeType, postCommand)) # node = pm.shadingNode(typ, asTexture=True)

		if not place2dTexture:
			pm.delete(pm.listConnections(node, type='place2dTexture', source=True, exactType=True))

		if tex:
			try:
				node.fileTextureName.set(tex)
			except Exception as error:
				print ('# Error:', __file__, error, '#')

		if name:
			try:
				pm.rename(node, name)

			except RuntimeError as error:
				print ('# Error:', __file__, error, '#')

		cls.setAttributesMEL(node, **kwargs)
		return node


	@staticmethod
	def getIncomingNodeByType(node, typ, exact=True):
		'''Get the first connected node of the given type with an incoming connection to the given node.

		:Parameters:
			node (str)(obj) = A node with incoming connections.
			typ (str) = The node type to search for. ie. 'StingrayPBS'
			exact (bool) = Only consider nodes of the exact type. Otherwise, derived types are also taken into account.

		:Return:
			(obj)(None) node if found.

		ex. call: env_file_node = getIncomingNodeByType(env_node, 'file') #get the incoming file node.
		'''
		nodes = pm.listConnections(node, type=typ, source=True, exactType=exact)
		return Iter.formatReturn([pm.PyNode(n) for n in nodes])


	@staticmethod
	def getOutgoingNodeByType(node, typ, exact=True):
		'''Get the connected node of the given type with an outgoing connection to the given node.

		:Parameters:
			node (str)(obj) = A node with outgoing connections.
			typ (str) = The node type to search for. ie. 'file'
			exact (bool) = Only consider nodes of the exact type. Otherwise, derived types are also taken into account.

		:Return:
			(list)(obj)(None) node(s)

		ex. call: srSG_node = getOutgoingNodeByType(sr_node, 'shadingEngine') #get the outgoing shadingEngine node.
		'''
		nodes = pm.listConnections(node, type=typ, destination=True, exactType=exact)
		return Iter.formatReturn([pm.PyNode(n) for n in nodes])


	@staticmethod
	def getAttributesMEL(node, inc=[], exc=[]):
		'''Get node attributes and their corresponding values as a dict.

		:Parameters:
			node (obj) = The node to get attributes for.
			inc (list) = Attributes to include. All other will be omitted. Exclude takes dominance over include. Meaning, if the same attribute is in both lists, it will be excluded.
			exc (list) = Attributes to exclude from the returned dictionay. ie. ['Position','Rotation','Scale','renderable','isHidden','isFrozen','selected']

		:Return:
			(dict) {'string attribute': current value}
		'''
		# print('node:', node); print('attr:', , read=1, hasData=1))
		attributes={}
		for attr in Iter.filterList(pm.listAttr(node, read=1, hasData=1), inc, exc):
			try:
				attributes[attr] = pm.getAttr(getattr(node, attr), silent=True) #get the attribute's value.
			except Exception as error:
				print (error)

		return attributes


	@staticmethod
	def setAttributesMEL(nodes, verbose=False, **attributes):
		'''Set node attribute values.

		:Parameters:
			nodes (str)(obj)(list) = The node(s) to set attributes for.
			verbose (bool) = Print feedback messages such as errors to the console.
			attributes (dict) = Attributes and their correponding value to set. ie. {'string attribute': value}

		ex call: setAttributesMEL(node, translateY=6)
		'''
		for node in pm.ls(nodes):

			for attr, value in attributes.items():
				try:
					a = getattr(node, attr)
					pm.setAttr(a, value) #ie. pm.setAttr('polyCube1.subdivisionsDepth', 5)
				except (AttributeError, TypeError) as error:
					if verbose:
						print ('# Error:', __file__, error, node, attr, value, '#')
					pass


	@staticmethod
	def connectAttributes(attr, place, file):
		'''A convenience procedure for connecting common attributes between two nodes.

		:Parameters:
			attr () = 
			place () = 
			file () = 

		ex. call:
		Use convenience command to connect attributes which share 
		their names for both the placement and file nodes.
			connectAttributes('coverage', 'place2d', fileNode')
			connectAttributes('translateFrame', 'place2d', fileNode')
			connectAttributes('rotateFrame', 'place2d', fileNode')
			connectAttributes('mirror', 'place2d', fileNode')
			connectAttributes('stagger', 'place2d', fileNode')
			connectAttributes('wrap', 'place2d', fileNode')
			connectAttributes('wrapV', 'place2d', fileNode')
			connectAttributes('repeatUV', 'place2d', fileNode')
			connectAttributes('offset', 'place2d', fileNode')
			connectAttributes('rotateUV', 'place2d', fileNode')

		These two are named differently.
			connectAttr -f ( $place2d + ".outUV" ) ( $fileNode + ".uv" );
			connectAttr -f ( $place2d + ".outUvFilterSize" ) ( $fileNode + ".uvFilterSize" );
		'''
		pm.connectAttr('{}.{}'.format(place, attr), '{}.{}'.format(file, attr), f=1)


	@staticmethod
	def connectMultiAttr(*args, force=True):
		'''Connect multiple node attributes at once.

		:Parameters:
			args (tuple) = Attributes as two element tuples. ie. (<connect from attribute>, <connect to attribute>)

		ex. call: connectMultiAttr(
			(node1.outColor, node2.aiSurfaceShader),
			(node1.outColor, node3.baseColor),
			(node4.outNormal, node5.normalCamera),
		)
		'''
		for frm, to in args:
			try:
				pm.connectAttr(frm, to)
			except Exception as error:
				print ('# Error:', __file__, error, '#')

# --------------------------------------------------------------------------------------------

def __getattr__(attr):
	'''Attempt to get a class attribute.

	:Return:
		(obj)
	'''
	try:
		return getattr(Node, attr)
	except AttributeError as error:
		raise AttributeError(f'{__file__} in __getattr__\n\t{error} ({type(attr).__name__})')







# print (__name__) #module name
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# deprecated: -----------------------------------


# def filterComponents(cls, frm, inc=[], exc=[]):
# 		'''Filter the given 'frm' list for the items in 'exc'.

# 		:Parameters:
# 			frm (str)(obj)(list) = The components(s) to filter.
# 			inc (str)(obj)(list) = The component(s) to include.
# 			exc (str)(obj)(list) = The component(s) to exclude.
# 								(exlude take precidence over include)
# 		:Return:
# 			(list)

# 		ex. call: filterComponents('obj.vtx[:]', 'obj.vtx[1:23]') #returns: [MeshVertex('objShape.vtx[0]'), MeshVertex('objShape.vtx[24]'), MeshVertex('objShape.vtx[25]')]
# 		'''
# 		exc = pm.ls(exc, flatten=True)
# 		if not exc:
# 			return frm

# 		c, *other = components = pm.ls(frm, flatten=True)
# 		#determine the type of items in 'exc' by sampling the first element.
# 		if isinstance(c, str):
# 			if 'Shape' in c:
# 				rtn = 'transform'
# 			else:
# 				rtn = 'str'
# 		elif isinstance(c, int):
# 			rtn = 'int'
# 		else:
# 			rtn = 'obj'

# 		if exc and isinstance(exc[0], int): #attempt to create a component list from the given integers. warning: this will only exclude from a single object.
# 			obj = pm.ls(frm, objectsOnly=1)
# 			if len(obj)>1:
# 				return frm
# 			componentType = cls.getComponentType(frm[0])
# 			typ = cls.convertAlias(componentType) #get the correct componentType variable from possible args.
# 			exc = ["{}.{}[{}]".format(obj[0], typ, n) for n in exc]

# 		if inc and isinstance(inc[0], int): #attempt to create a component list from the given integers. warning: this will only exclude from a single object.
# 			obj = pm.ls(frm, objectsOnly=1)
# 			if len(obj)>1:
# 				return frm
# 			componentType = cls.getComponentType(frm[0])
# 			typ = cls.convertAlias(componentType) #get the correct componentType variable from possible args.
# 			inc = ["{}.{}[{}]".format(obj[0], typ, n) for n in inc]

# 		inc = convertArrayType(inc, returnType=rtn, flatten=True) #assure both lists are of the same type for comparison.
# 		exc = convertArrayType(exc, returnType=rtn, flatten=True)
# 		return [i for i in components if i not in exc and (inc and i in inc)]


# !/usr/bin/python
# coding=utf-8
try:
	import pymel.core as pm
except ImportError as error:
	print (__file__, error)

from pythontk import Iter, Math
#from this package:
from mayatk.core import mfnMeshGenerator, viewportMessage, undo
from mayatk.cmpttk import Cmpt


class Xform(object):
	'''
	'''
	@staticmethod
	def moveTo(source, target, targetCenter=True):
		'''Move an object(s) to the given target.

		:Parameters:
			source (str)(obj)(list) = The objects to move.
			target (str)(obj) = The object to move to.
			targetCenter (bool) = Move to target pivot pos, or the bounding box center of the target.
		'''
		if targetCenter: #temporarily move the targets pivot to it's bounding box center.
			orig_target_piv = pm.xform(target, q=1, worldSpace=1, rp=1) #get target pivot position.
			pm.xform(target, centerPivots=1) #center target pivot.
			target_pos = pm.xform(target, q=1, worldSpace=1, rp=1) #get the pivot position at center of object.
			pm.xform(target, worldSpace=1, rp=orig_target_piv) #return the target pivot to it's orig position.
		else:
			target_pos = pm.xform(target, q=1, worldSpace=1, rp=1) #get the pivot position.

		pm.xform(source, translation=target_pos, worldSpace=1, relative=1)


	@staticmethod
	@undo
	def dropToGrid(objects, align='Mid', origin=False, centerPivot=False, freezeTransforms=False):
		'''Align objects to Y origin on the grid using a helper plane.

		:Parameters:
			objects (str)(obj)(list) = The objects to translate.
			align (bool) = Specify which point of the object's bounding box to align with the grid. (valid: 'Max','Mid'(default),'Min')
			origin (bool) = Move to world grid's center.
			centerPivot (bool) = Center the object's pivot.
			freezeTransforms (bool) = Reset the selected transform and all of its children down to the shape level.

		ex. dropToGrid(obj, align='Min') #set the object onto the grid.
		'''
		# pm.undoInfo(openChunk=1)
		for obj in pm.ls(objects, transforms=1):
			osPivot = pm.xform(obj, query=1, rotatePivot=1, objectSpace=1) #save the object space obj pivot.
			wsPivot = pm.xform(obj, query=1, rotatePivot=1, worldSpace=1) #save the world space obj pivot.

			pm.xform(obj, centerPivots=1) #center pivot
			plane = pm.polyPlane(name='temp#')

			if not origin:
				pm.xform(plane, translation=(wsPivot[0], 0, wsPivot[2]), absolute=1, ws=1) #move the object to the pivot location

			pm.align(obj, plane, atl=1, x='Mid', y=align, z='Mid')
			pm.delete(plane)

			if not centerPivot:
				pm.xform(obj, rotatePivot=osPivot, objectSpace=1) #return pivot to orig position.

			if freezeTransforms:
				pm.makeIdentity(obj, apply=True)
		# pm.undoInfo (closeChunk=1)


	@classmethod
	@undo
	def resetTranslation(cls, objects):
		'''Reset the translation transformations on the given object(s).

		:Parameters:
			objects (str)(obj)(list) = The object(s) to reset the translation values for.
		'''
		# pm.undoInfo(openChunk=1)
		for obj in pm.ls(objects):
			pos = pm.objectCenter(obj) #get the object's current position.
			cls.dropToGrid(obj, origin=1, centerPivot=1) #move to origin and center pivot.
			pm.makeIdentity(obj, apply=1, t=1, r=0, s=0, n=0, pn=1) #bake transforms
			pm.xform(obj, translation=pos) #move the object back to it's original position.
		# pm.undoInfo(closeChunk=1)


	@staticmethod
	def setTranslationToPivot(node):
		'''Set an object’s translation value from its pivot location.

		:Parameters:
			node (str)(obj)(list) = An object, or it's name.
		'''
		x, y, z = pivot = pm.xform(node, query=True, worldSpace=True, rotatePivot=True)
		pm.xform(node, relative=True, translation=[-x,-y,-z])
		pm.makeIdentity(node, apply=True, translate=True)
		pm.xform(node, translation=[x, y, z])


	@staticmethod
	@undo
	def alignPivotToSelection(alignFrom=[], alignTo=[], translate=True):
		'''Align one objects pivot point to another using 3 point align.

		:Parameters:
			alignFrom (list) = At minimum; 1 object, 1 Face, 2 Edges, or 3 Vertices.
			alignTo (list) = The object to align with.
			translate (bool) = Move the object with it's pivot.
		'''
		# pm.undoInfo(openChunk=1)
		pos = pm.xform(alignTo, q=1, translation=True, worldSpace=True)
		center_pos = [ #Get center by averaging of all x,y,z points.
			sum(pos[0::3]) / len(pos[0::3]), 
			sum(pos[1::3]) / len(pos[1::3]), 
			sum(pos[2::3]) / len(pos[2::3])]

		vertices = pm.ls(pm.polyListComponentConversion(alignTo, toVertex=True), flatten=True)
		if len(vertices) < 3:
			return

		for obj in pm.ls(alignFrom, flatten=1):

			plane = pm.polyPlane(name="_hptemp#", width=1, height=1, subdivisionsX=1, subdivisionsY=1, axis=[0, 1, 0], createUVs=2, constructionHistory=True)[0] #Create and align helper plane.

			pm.select("%s.vtx[0:2]" % plane, vertices[0:3])
			pm.mel.snap3PointsTo3Points(0)

			pm.xform(obj, rotation=pm.xform(plane, q=True, rotation=True, worldSpace=True), worldSpace=True)

			if translate:
				pm.xform(obj, translation=center_pos, worldSpace=True)
				
			pm.delete(plane)
		# pm.undoInfo(closeChunk=1)


	@staticmethod
	@undo
	def aimObjectAtPoint(objects, target_pos, aim_vect=(1,0,0), up_vect=(0,1,0)):
		'''Aim the given object(s) at the given world space position.

		:Paramters:
			objects (str)(obj)(list) = Transform node(s) of the objects to orient.
			target_pos (obj)(tuple) = A point as xyz, or one or more transform nodes at which to aim the other given 'objects'.
			aim_vect (tuple) = The vector in local coordinates that points at the target.
			up_vect (tuple) = The vector in local coordinates that aligns with the world up vector.

		ex. call: aimObjectAtPoint(['cube1', 'cube2'], (0, 15, 15))
		'''
		if isinstance(target_pos, (tuple, set, list)):
			target = pm.createNode('transform', name='target_helper')

		pm.xform(target, translation=target_pos, absolute=True)

		for obj in Iter.makeList(objects):
			const = pm.aimConstraint(target, obj, aim=aim_vect, worldUpVector=up_vect, worldUpType="vector")

		pm.delete(const, target)


	@classmethod
	@undo
	def rotateAxis(cls, objects, target_pos):
		''' Aim the given object at the given world space position.
		All rotations in rotated channel, geometry is transformed so 
		it does not appear to move during this transformation

		:Parameters:
			objects (str)(obj)(list) = Transform node(s) of the objects to orient.
			target_pos (obj)(tuple) = A point as xyz, or one or more transform nodes at which to aim the other given 'objects'.
		'''
		for obj in pm.ls(objects, objectsOnly=True):
			cls.aimObjectAtPoint(obj, target_pos)

			try:
				c = obj.verts
			except TypeError as error:
				c = obj.cp

			wim = pm.getAttr(obj.worldInverseMatrix)
			pm.xform(c, matrix=wim)

			pos = pm.xform(obj, q=True, translation=True, absolute=True, worldSpace=True)
			pm.xform(c, translation=pos, relative=True, worldSpace=True)


	@staticmethod
	def getOrientation(objects, returnType='point'):
		'''Get an objects orientation as a point or vector.

		:Parameters:
			objects (str)(obj)(list) = The object(s) to get the orientation of.
			returnType (str) = The desired returned value type. (valid: 'point'(default), 'vector')

		:Return:
			(tuple)(list) If 'objects' given as a list, a list of tuples will be returned.
		'''
		result=[]
		for obj in pm.ls(objects, objectsOnly=True):

			world_matrix = pm.xform(obj, q=True, matrix=True, worldSpace=True)
			rAxis = pm.getAttr(obj.rotateAxis)
			if any((rAxis[0], rAxis[1], rAxis[2])):
				print('# Warning: {} has a modified .rotateAxis of {} which is included in the result. #'.format(obj, rAxis))

			if returnType=='vector':
				from maya.api.OpenMaya import MVector
				ori = (
					MVector(world_matrix[0:3]),
					MVector(world_matrix[4:7]),
					MVector(world_matrix[8:11])
				)

			else:
				ori = (
					world_matrix[0:3],
					world_matrix[4:7],
					world_matrix[8:11]
				)
			result.append(ori)

		return Iter.formatReturn(result, objects)


	@staticmethod
	def getDistanceBetweenTwoObjects(objA, objB):
		'''Get the magnatude of a vector using the center points of two given objects.

		:Parameters:
			objA (obj)(str) = Object, object name, or point (x,y,z).
			objB (obj)(str) = Object, object name, or point (x,y,z).

		:Return:
			(float)

		# xmin, ymin, zmin, xmax, ymax, zmax = pm.exactWorldBoundingBox(startAndEndCurves)
		'''
		x1, y1, z1 = pm.objectCenter(objA)
		x2, y2, z2 = pm.objectCenter(objB)

		from math import sqrt
		distance = sqrt(pow((x1-x2),2) + pow((y1-y2),2) + pow((z1-z2),2))

		return distance


	@staticmethod
	def getCenterPoint(objects):
		'''Get the bounding box center point of any given object(s).
		
		:Parameters:
			objects (str)(obj(list) = The objects or components to get the center of.

		:Return:
			(tuple) position as xyz float values.
		'''
		objects = pm.ls(objects, flatten=True)
		pos = [i for sublist in [pm.xform(s, q=1, translation=1, worldSpace=1, absolute=1) for s in objects] for i in sublist]
		center_pos = ( #Get center by averaging of all x,y,z points.
			sum(pos[0::3]) / len(pos[0::3]), 
			sum(pos[1::3]) / len(pos[1::3]), 
			sum(pos[2::3]) / len(pos[2::3])
		)
		return center_pos


	@classmethod
	def getBoundingBoxValue(cls, objects, value='sizeX|sizeY|sizeZ'):
		'''Get information of the given object(s) combined bounding box.

		:Parameters:
			objects (str)(obj)(list) = The object(s) or component(s) to query.
				Multiple objects will be treated as a combined bounding box.
			value (str) = The type of value to return. Multiple types can be given
				separated by '|'. The order given determines the return order.
				valid (case insensitive): 'xmin', 'xmax', 'ymin', 'ymax', 
				'zmin', 'zmax', 'sizex', 'sizey', 'sizez', 'volume', 'center'

		:Return:
			(float)(tuple) Dependant on args.

		ex. call: getBoundingBoxValue(sel, 'center|volume') #returns: [[171.9106216430664, 93.622802734375, -1308.4896240234375], 743.2855185396038]
		ex. call: getBoundingBoxValue(sel, 'sizeY') #returns: 144.71902465820312
		'''
		if '|' in value: #use recursion to construct the list using each value.
			return tuple(cls.getBoundingBoxValue(objects, i) for i in value.split('|'))

		v = value.lower()
		xmin, ymin, zmin, xmax, ymax, zmax = pm.exactWorldBoundingBox(objects)
		if v=='xmin': return xmin
		elif v=='xmax': return xmax
		elif v=='ymin': return ymin
		elif v=='ymax': return ymax
		elif v=='zmin': return zmin
		elif v=='zmax': return zmax
		elif v=='sizex': return xmax-xmin
		elif v=='sizey': return ymax-ymin
		elif v=='sizez': return zmax-zmin
		elif v=='volume': return (xmax-xmin)*(ymax-ymin)*(zmax-zmin)
		elif v=='center': return ((xmin+xmax)/2.0, (ymin+ymax)/2.0, (zmin+zmax)/2.0)


	@classmethod
	def sortByBoundingBoxValue(cls, objects, value='volume', descending=True, returnWithValue=False):
		'''Sort the given objects by their bounding box value.

		:Parameters:
			objects (str)(obj)(list) = The objects or components to sort.
			value (str) = See 'getBoundingBoxInfo' 'value' parameter.
					ex. 'xmin', 'xmax', 'sizex', 'volume', 'center' etc.
			descending (bool) = Sort the list from the largest value down.
			returnWithValue (bool) = Instead of just the object; return a 
					list of two element tuples as [(<value>, <obj>)].
		:Return:
			(list)
		'''
		valueAndObjs=[]
		for obj in pm.ls(objects, flatten=False):
			v = cls.getBoundingBoxValue(obj, value)
			print (0, obj, v)
			valueAndObjs.append((v, obj))

		sorted_ = sorted(valueAndObjs, key=lambda x: int(x[0]), reverse=descending)
		if returnWithValue:
			return sorted_
		return [obj for v, obj in sorted_]


	@staticmethod
	def matchScale(objectsA, objectsB, scale=True, average=False):
		'''Scale each of the given objects to the combined bounding box of a second set of objects.

		:Parameters:
			objectsA (str)(obj)(list) = The object(s) to scale.
			objectsB (str)(obj)(list) = The object(s) to get a bounding box size from.
			scale (bool) = Scale the objects. Else, just return the scale value.
			average (bool) = Average the result across all axes.

		:Return:
			(list) scale values as [x,y,z,x,y,z...]
		'''
		to = pm.ls(objectsA, flatten=True)
		frm = pm.ls(objectsB, flatten=True)

		xmin, ymin, zmin, xmax, ymax, zmax = pm.exactWorldBoundingBox(frm)
		ax, ay, az = aBoundBox = [xmax-xmin, ymax-ymin, zmax-zmin]

		result=[]
		for obj in to:

			xmin, ymin, zmin, xmax, ymax, zmax = pm.exactWorldBoundingBox(obj)
			bx, by, bz = bBoundBox = [xmax-xmin, ymax-ymin, zmax-zmin]

			oldx, oldy, oldz = bScaleOld = pm.xform(obj, q=1, s=1, r=1)

			try:
				diffx, diffy, diffz = boundDifference = [ax/bx, ay/by, az/bz]
			except ZeroDivisionError as error:
				diffx, diffy, diffz = boundDifference = [1, 1, 1]

			bScaleNew = [oldx*diffx, oldy*diffy, oldz*diffz]

			if average:
				bScaleNew = [sum(bScaleNew)/len(bScaleNew) for _ in range(3)]

			if scale:
				pm.xform(obj, scale=bScaleNew)

			[result.append(i) for i in bScaleNew]

		return result


	@staticmethod
	def snap3PointsTo3Points(vertices):
		'''Move and align the object defined by the first 3 points to the last 3 points.

		:Parameters:
			vertices (list) = The first 3 points must be on the same object (i.e. it is the 
						object to be transformed). The second set of points define 
						the position and plane to transform to.
		'''
		import math

		vertices = pm.ls(vertices, flatten=True)
		objectToMove = pm.ls(vertices[:3], objectsOnly=True)

		# get the world space position of each selected point object
		p0, p1, p2 = [pm.pointPosition(v) for v in vertices[0:3]]
		p3, p4, p5 = [pm.pointPosition(v) for v in vertices[3:6]]

		dx, dy, dz = distance = [ # calculate the translation amount - the first point on each pair is the point to use for translation.
			p3[0] - p0[0],
			p3[1] - p0[1],
			p3[2] - p0[2]
		]

		pm.move(dx, dy, dz, objectToMove, relative=1) # move the first object by that amount.

		a1x, a1y, a1z = axis1 = [ # define the two vectors for each pair of points.
			p1[0] - p0[0],
			p1[1] - p0[1],
			p1[2] - p0[2]
		]
		a2x, a2y, a2z = axis2 = [
			p4[0] - p3[0],
			p4[1] - p3[1],
			p4[2] - p3[2]
		]

		# get the angle (in radians) between the two vectors and the axis of rotation. This is used to move axis1 to match axis2
		dp = Math.dotProduct(axis1, axis2, 1)
		dp = Math.clamp(-1.0, 1.0, dp)
		angle = math.acos(dp)
		crossProduct = Math.crossProduct(axis1, axis2, 1, 1)

	 	# rotate the first object about the pivot point (the pivot is defined by the first point from the second pair of points. i.e. point 3 from the inputs above)
		rotation = Math.xyzRotation(angle, crossProduct)
		pm.rotate(objectToMove, str(rotation[0])+'rad', str(rotation[1])+'rad', str(rotation[2])+'rad', pivot=p3, relative=1)

		# Get these points again since they may have moved
		p2 = pm.pointPosition(vertices[2])
		p5 = pm.pointPosition(vertices[5])

		axis3 = [
			p2[0] - p4[0],
			p2[1] - p4[1],
			p2[2] - p4[2]
		]
		axis4 = [
			p5[0] - p4[0],
			p5[1] - p4[1],
			p5[2] - p4[2]
		]

		axis2 = Math.normalize(axis2)

		# Get the dot product of axis3 on axis2
		dp = Math.dotProduct(axis3, axis2, 0)
		axis3[0] = p2[0] - p4[0] + dp * axis2[0]
		axis3[1] = p2[1] - p4[1] + dp * axis2[1]
		axis3[2] = p2[2] - p4[2] + dp * axis2[2]

		# Get the dot product of axis4 on axis2
		dp = Math.dotProduct(axis4, axis2, 0)
		axis4[0] = p5[0] - p4[0] + dp * axis2[0]
		axis4[1] = p5[1] - p4[1] + dp * axis2[1]
		axis4[2] = p5[2] - p4[2] + dp * axis2[2]

		# rotate the first object again, this time about the 2nd axis so that the 3rd point is in the same plane. ie. match up axis3 with axis4.
		dp = Math.dotProduct(axis3, axis4, 1)
		dp = Math.clamp(-1.0, 1.0, dp)
		angle = math.acos(dp)

		# reverse the angle if the cross product is in the -ve axis direction
		crossProduct = Math.crossProduct(axis3, axis4, 1, 1)
		dp = Math.dotProduct(crossProduct, axis2, 0)
		if dp < 0:
			angle = -angle

		rotation = Math.xyzRotation(angle, axis2)
		pm.rotate(objectToMove, str(rotation[0])+'rad', str(rotation[1])+'rad', str(rotation[2])+'rad', pivot=p4, relative=1)


	@staticmethod
	def isOverlapping(objA, objB, tolerance=0.001):
		'''Check if the vertices in objA and objB are overlapping within the given tolerance.

		:Parameters:
			objA (str)(obj) = The first object to check. Object can be a component.
			objB (str)(obj) = The second object to check. Object can be a component.
			tolerance (float) = The maximum search distance before a vertex is considered not overlapping.

		:Return:
			(bool)
		'''
		vert_setA = pm.ls(pm.polyListComponentConversion(objA, toVertex=1), flatten=1)
		vert_setB = pm.ls(pm.polyListComponentConversion(objB, toVertex=1), flatten=1)

		closestVerts = Cmpt.getClosestVerts(vert_setA, vert_setB, tolerance=tolerance)

		return True if vert_setA and len(closestVerts)==len(vert_setA) else False


	@staticmethod
	def getVertPositions(objects, worldSpace=True):
		'''Get all vertex positions for the given objects.

		:Parameters:
			objects (str)(obj)(list) = The polygon object(s).
			worldSpace (bool) = Sample in world or object space.

		:Return:
			(list) Nested lists if multiple objects given.
		'''
		import maya.OpenMaya as om
		space = om.MSpace.kWorld if worldSpace else om.MSpace.kObject

		result=[]
		for mesh in mfnMeshGenerator(objects):
			points = om.MPointArray()
			mesh.getPoints(points, space)

			result.append([(points[i][0], points[i][1], points[i][2]) for i in range(points.length())])
		return Iter.formatReturn(result, objects)


	@staticmethod
	def hashPoints(points, precision=4):
		'''Hash the given list of point values.

		:Parameters:
			points (list) = A list of point values as tuples.
			precision (int) = determines the number of decimal places that are retained 
				in the fixed-point representation. For example, with a value of 4, the 
				fixed-point representation would retain 4 decimal place.

		:Return:
			(list) list(s) of hashed tuples.
		'''
		nested = Iter.nestedDepth(points)>1
		sets = points if nested else [points]

		result=[]
		for pset in sets:
			clamp = lambda p: int(p * 10**precision)

			result.append([hash(tuple(map(clamp, i))) for i in pset])
		return Iter.formatReturn(result, nested)


	@classmethod
	def getMatchingVerts(cls, meshA, meshB, worldSpace=False):
		'''Find any vertices which point locations match between two given mesh.

		:Parameters:
			meshA (str)(obj)(list) = The first polygon object.
			meshA (str)(obj)(list) = A second polygon object.
			worldSpace (bool) = Sample in world or object space.

		:Return:
			(list) nested tuples with int values representing matching vertex pairs.
		'''
		vertPosA, vertPosB = cls.getVertPositions([meshA, meshB], worldSpace)
		hashA, hashB = cls.hashPoints([vertPosA, vertPosB])

		matching = set(hashA).intersection(hashB)
		return [i for h in matching 
			for i in zip(Iter.indices(hashA, h), Iter.indices(hashB, h))]


	@staticmethod
	def orderByDistance(objects, point=[0, 0, 0], reverse=False):
		'''Order the given objects by their distance from the given point.
		:Parameters:
			objects (str)(int)(list) = The object(s) to order.
			point (list) = A three value float list x, y, z.
			reverse (bool) = Reverse the naming order. (Farthest object first)
		:Return:
			(list) ordered objects
		'''
		distance={}
		for obj in pm.ls(objects, flatten=1):
			xmin, ymin, zmin, xmax, ymax, zmax = pm.xform(obj, q=1, boundingBox=1)
			bb_pos = ((xmin + xmax) / 2, (ymin + ymax) / 2, (zmin + zmax) / 2)
			bb_dist = Math.getDistBetweenTwoPoints(point, bb_pos)

			distance[bb_dist] = obj

		result = [distance[i] for i in sorted(distance)]
		return list(reversed(result)) if reverse else result


	@staticmethod
	@undo
	def alignVertices(mode, average=False, edgeloop=False):
		'''Align vertices.

		:Parameters:
			mode (int) = possible values are align: 0-YZ, 1-XZ, 2-XY, 3-X, 4-Y, 5-Z, 6-XYZ 
			average (bool) = align to average of all selected vertices. else, align to last selected
			edgeloop (bool) = align vertices in edgeloop from a selected edge

		ex. call: alignVertices(mode=3, average=True, edgeloop=True)
		'''
		# pm.undoInfo (openChunk=True)
		selectTypeEdge = pm.selectType(query=True, edge=True)

		if edgeloop:
			pm.mel.SelectEdgeLoopSp() #select edgeloop

		pm.mel.PolySelectConvert(3) #convert to vertices

		selection = pm.ls(selection=1, flatten=1)
		lastSelected = pm.ls(tail=1, selection=1, flatten=1)
		alignTo = pm.xform(lastSelected, query=1, translation=1, worldSpace=1)
		alignX = alignTo[0]
		alignY = alignTo[1]
		alignZ = alignTo[2]
		
		if average:
			xyz = pm.xform(selection, query=1, translation=1, worldSpace=1)
			x = xyz[0::3]
			y = xyz[1::3]
			z = xyz[2::3]
			alignX = float(sum(x))/(len(xyz)/3)
			alignY = float(sum(y))/(len(xyz)/3)
			alignZ = float(sum(z))/(len(xyz)/3)

		if len(selection)<2:
			if len(selection)==0:
				viewportMessage("No vertices selected")
			viewportMessage("Selection must contain at least two vertices")

		for vertex in selection:
			vertexXYZ = pm.xform(vertex, query=1, translation=1, worldSpace=1)
			vertX = vertexXYZ[0]
			vertY = vertexXYZ[1]
			vertZ = vertexXYZ[2]
			
			modes = {
				0:(vertX, alignY, alignZ), #align YZ
				1:(alignX, vertY, alignZ), #align XZ
				2:(alignX, alignY, vertZ), #align XY
				3:(alignX, vertY, vertZ),
				4:(vertX, alignY, vertZ),
				5:(vertX, vertY, alignZ),
				6:(alignX, alignY, alignZ), #align XYZ
			}

			pm.xform(vertex, translation=modes[mode], worldSpace=1)

		if selectTypeEdge:
			pm.selectType (edge=True)
		# pm.undoInfo (closeChunk=True)


	

# --------------------------------------------------------------------------------------------

def __getattr__(attr):
	'''Attempt to get a class attribute.

	:Return:
		(obj)
	'''
	try:
		return getattr(Xform, attr)
	except AttributeError as error:
		raise AttributeError(f'{__file__} in __getattr__\n\t{error} ({type(attr).__name__})')







# --------------------------------------------------------------------------------------------

if __name__=='__main__':
	pass


# --------------------------------------------------------------------------------------------









# print (__name__) #module name
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# deprecated: -----------------------------------

# @classmethod
# 	def matchTransformByVertexOrder(cls, source, target):
# 		'''Match transform and rotation on like objects by using 3 vertices from each object.
# 		The vertex order is transferred to the target object(s).

# 		:Parameters:
# 			source (str)(obj) = The object to move from.
# 			target (str)(obj) = The object to move to.
# 		'''
# 		pm.polyTransfer(source, alternateObject=target, vertices=2) #vertices positions are copied from the target object.

# 		source_verts = [pm.ls(source, objectsOnly=1)[0].verts[i] for i in range(3)]
# 		target_verts = [pm.ls(target, objectsOnly=1)[0].verts[i] for i in range(3)]

# 		cls.snap3PointsTo3Points(source_verts+target_verts)

# @staticmethod
# 	def getComponentPoint(component, alignToNormal=False):
# 		'''Get the center point from the given component.

# 		:Parameters:
# 			component (str)(obj) = Object component.
# 			alignToNormal (bool) = Constain to normal vector.

# 		:Return:
# 			(tuple) coordinate as xyz float values.
# 		'''
# 		if ".vtx" in str(component):
# 			x = pm.polyNormalPerVertex(component, query=1, x=1)
# 			y = pm.polyNormalPerVertex(component, query=1, y=1)
# 			z = pm.polyNormalPerVertex(component, query=1, z=1)
# 			xyz = [sum(x) / float(len(x)), sum(y) / float(len(y)), sum(z) / float(len(z))] #get average

# 		elif ".e" in str(component):
# 			componentName = str(component).split(".")[0]
# 			vertices = pm.polyInfo (component, edgeToVertex=1)[0]
# 			vertices = vertices.split()
# 			vertices = [componentName+".vtx["+vertices[2]+"]",componentName+".vtx["+vertices[3]+"]"]
# 			x=[];y=[];z=[]
# 			for vertex in vertices:
# 				x_ = pm.polyNormalPerVertex (vertex, query=1, x=1)
# 				x.append(sum(x_) / float(len(x_)))
# 				y_ = pm.polyNormalPerVertex (vertex, query=1, y=1)
# 				x.append(sum(y_) / float(len(y_)))
# 				z_ = pm.polyNormalPerVertex (vertex, query=1, z=1)
# 				x.append(sum(z_) / float(len(z_)))
# 			xyz = [sum(x) / float(len(x)), sum(y) / float(len(y)), sum(z) / float(len(z))] #get average

# 		else:# elif ".f" in str(component):
# 			xyz = pm.polyInfo (component, faceNormals=1)
# 			xyz = xyz[0].split()
# 			xyz = [float(xyz[2]), float(xyz[3]), float(xyz[4])]

# 		if alignToNormal: #normal constraint
# 			normal = pm.mel.eval("unit <<"+str(xyz[0])+", "+str(xyz[1])+", "+str(xyz[2])+">>;") #normalize value using MEL
# 			# normal = [round(i-min(xyz)/(max(xyz)-min(xyz)),6) for i in xyz] #normalize and round value using python

# 			constraint = pm.normalConstraint(component, object_,aimVector=normal,upVector=[0,1,0],worldUpVector=[0,1,0],worldUpType="vector") # "scene","object","objectrotation","vector","none"
# 			pm.delete(constraint) #orient object_ then remove constraint.

# 		vertexPoint = pm.xform (component, query=1, translation=1) #average vertex points on destination to get component center.
# 		x = vertexPoint[0::3]
# 		y = vertexPoint[1::3]
# 		z = vertexPoint[2::3]

# 		return tuple(round(sum(x) / float(len(x)),4), round(sum(y) / float(len(y)),4), round(sum(z) / float(len(z)),4))
